From 8178766bddd837efc8608d3b05d3213e690c7250 Mon Sep 17 00:00:00 2001
From: Thomas Munro <thomas.munro@enterprisedb.com>
Date: Sat, 12 Jan 2019 02:17:02 +1300
Subject: [PATCH 04/14] Allow WAL record data on first modification after a
 checkpoint.

Provide a way to attach data to WAL record conditionally, so that it is
included only if this turns out to the be first modification to a given
block after a checkpoint.

This will be used to record undo log meta-data, to avoid a data
synchronization problem with online checkpoints.

Author: Thomas Munro
Reviewed-by:
Discussion:
---
 src/backend/access/transam/xloginsert.c | 15 +++++++++++++++
 src/include/access/xloginsert.h         |  2 ++
 2 files changed, 17 insertions(+)

diff --git a/src/backend/access/transam/xloginsert.c b/src/backend/access/transam/xloginsert.c
index 1697797..36bf458 100644
--- a/src/backend/access/transam/xloginsert.c
+++ b/src/backend/access/transam/xloginsert.c
@@ -571,6 +571,21 @@ XLogRecordAssemble(RmgrId rmid, uint8 info,
 			needs_data = false;
 		else if ((regbuf->flags & REGBUF_KEEP_DATA) != 0)
 			needs_data = true;
+		else if ((regbuf->flags & REGBUF_KEEP_DATA_AFTER_CP) != 0)
+		{
+			XLogRecPtr	page_lsn = PageGetLSN(regbuf->page);
+
+			needs_data = (page_lsn <= RedoRecPtr);
+			if (!needs_data)
+			{
+				/*
+				 * XLogInsertRecord() will detect if our view of the latest
+				 * checkpoint's RedoRecPtr is out of date.
+				 */
+				if (*fpw_lsn == InvalidXLogRecPtr || page_lsn < *fpw_lsn)
+					*fpw_lsn = page_lsn;
+			}
+		}
 		else
 			needs_data = !needs_backup;
 
diff --git a/src/include/access/xloginsert.h b/src/include/access/xloginsert.h
index 3ae9d22..616f553 100644
--- a/src/include/access/xloginsert.h
+++ b/src/include/access/xloginsert.h
@@ -38,6 +38,8 @@
 									 * will be skipped) */
 #define REGBUF_KEEP_DATA	0x10	/* include data even if a full-page image
 									 * is taken */
+#define REGBUF_KEEP_DATA_AFTER_CP 0x20 /* include data on the first
+										* modification after a checkpoint */
 
 /* prototypes for public functions in xloginsert.c: */
 extern void XLogBeginInsert(void);
-- 
1.8.3.1

